(* ::Package:: *)

BeginPackage["MusicGenerator`"];


Render\[Breve]Music::usage="Music\[Breve]Render[music] \[Dash] plays a music string";


Parse\[Breve]Music::usage="Parse\[Breve]Music[expr] \[Dash] Parses a music string to a Mathematica expression";


Set\[Breve]Instruments::usage="";


seq::usage="Represents two or more notes to be played sequentially";


par::usage="Represents two or more notes to be played together";


Import\[Breve]MusicXML::usage="Import\[Breve]musicXML[file] \[Dash] Imports an XML file containing musicXML";


To\[Breve]Music\[Breve]String::usage="Converts a music expressio to a music string";


(*Note\[Breve]Pattern::usage="Note\[Breve]Pattern[] \[Dash] returns a pattern that matches an actual note (as opposed to a rest or directive) in a music expression. It can't be used with a music string";*)


Note\[Breve]StringQ::usage="Note\[Breve]StringQ[x]\[Dash] tests a string to determine if it matches an actual note (as opposed to a rest or directive) in a music expression. It can't be used with a music string";


Add\[Breve]Synthetic\[Breve]Sound::usage="";


Frequency\[Breve]From\[Breve]Note::usage="Frequency\[Breve]From\[Breve]Note.....";


Set\[Breve]Time\[Breve]Rate\[Breve]Function::usage="Set\[Breve]Time\[Breve]Rate\[Breve]Function[fn] \[Dash] Specifies a function to be called (with no arguments) when the time map has been completed. The function can modify the time map.";


Time\[Breve]Rate::usage="Time\[Breve]Rate[v] \[Dash] Returns the time rate function expressed as a function of x";


Set\[Breve]Time\[Breve]Rate::usage="Set\[Breve]Time\[Breve]Rate[v] \[Dash] Sets the  time rate function expressed as a function of x";


Transpose\[Breve]Note::usage="";


Invert\[Breve]Note\[Breve]About::usage="";


MusicGeneratorVersion[]="1.1";


{rep,trp,trm,rall(*,bar*),Bar};


{add9,dim7,min9,maj9,dom9,min6,maj6,sus2,sus4,min7,maj7,dom7,dim,aug,min,maj};


Begin["`Private`"];


Debug\[Breve]Print=SuperWidgetPackage`Debug\[Breve]Print;




(* ::Section:: *)
(*To Do*)


(* ::Text:: *)
(*\[FilledSmallCircle]  Sort out tempo definition, and add other time signatures*)
(*\[FilledSmallCircle]  Define note styles*)
(*\[FilledSmallCircle]  Consider interaction with music as it is played*)
(*\[FilledSmallCircle]  Crescendo and similar functions*)
(*\[FilledSmallCircle]  Decent synthetic sounds*)


(* ::Section:: *)
(*Parse\[Breve]Music*)


String\[Breve]Replace\[Breve]All[s_String,reps_]:=Module[{last=s,ans},
   ans=StringReplace[last,reps];
   While[last=!=ans,
      last=ans;
      ans=StringReplace[last,reps];
   ];
   ans
];

Set\[Breve]Instruments[ll_List]:=(instruments=ll;);

(* Note that at the moment we don't flatten seq constructs. This avoids picking out those containing directives the require resetting *)
Parse\[Breve]Music[str_]:=Module[{s0,s1,s2,s3,s4,s5,s6,s7},
   s0=FromCharacterCode[Select[ToCharacterCode[str],(#<60000)&]];
   s0=StringReplace[s0,{Shortest["(*"~~___~~"*)"]->"","<"->" Vf ",">"->" Vp "}];
   s1=StringReplace[" ("<>s0<>")",{","->" \[Paragraph] ",")"->") ","'"->"\[Breve]"}];
   s2=String\[Breve]Replace\[Breve]All[s1,{" "->",", " "->",", "("->"[",  ")"->"]", x:Except[WordCharacter]~~"[":>x<>" seq["}];
   s3=String\[Breve]Replace\[Breve]All[s2,{",,"->",","[,"->"[",",]"->"]","+,"->"+",",+"->"+",StartOfString~~","->"",","~~EndOfString->""}];
   s4=StringReplace[s3,(x:(LetterCharacter~~Longest[(WordCharacter|"\[Flat]"|"\[Sharp]"|"\[Natural]"|"."|"*"|"/"|"\[Breve]")..]):>"\""<>x<>"\"")];
   s5=ToExpression[s4]//.s_String[x___]:>ToExpression["Hold["<>s<>"]"][x]/.Plus->par/.
                   Hold[seq]->seq;
    s6=s5//.Hold[fn_][a___,s:Longest[(_String|_seq|_par|Hold[_][___])..]/;Length[{s}]>1,b___]:>Hold[fn][a,seq[s],b];
    s7=s6 //. Hold[fn_][x___]:>fn\[Breve]resolve[fn,{x}]/. s_String:>StringReplace[s,"\[Breve]"->"'"]
];

fn\[Breve]resolve[fn_,args_]:=Module[{a1,s},
         a1=args//.{a___,s:Longest[(_String|_seq|_par|Hold[_][___])..]/;Length[{s}]>1,b___}:>{a,seq[s],b};
         a1=a1//.{a___,b_seq,c_seq,d___}:>{a,Append[b,c],d};
         a1=With[{x=ToExpression["\[Paragraph]"]},a1/.x:>Sequence@@{}];
         Hold[fn]@@a1
];
       


(* ::Section:: *)
(*Sundry functions*)


Error[mes_]:=(
   Print[StylePrint["Error: "<>mes,Background->Green]];
   Abort[]
);


SetAttributes[String\[Breve]Parse,HoldFirst]
String\[Breve]Parse[str_,pat_]:=Module[{x},
   parsed\[Breve]result=Null;
   ans=StringReplace[str,StartOfString~~(x:Longest[pat]~~___~~EndOfString):>(parsed\[Breve]result=x;"")];
   If[parsed\[Breve]result===Null,Return[False]];
   str=StringDrop[str,StringLength[parsed\[Breve]result]];
   True
];


(* ::Section:: *)
(*Process key signatures and note numbers*)


twelve\[Breve]notes={"C","C\[Sharp]","D","D\[Sharp]","E","F","F\[Sharp]","G","G\[Sharp]","A","A\[Sharp]","B"};
twelve\[Breve]notes1={"C","D\[Flat]","D","E\[Flat]","E","F","G\[Flat]","G","A\[Flat]","A","B\[Flat]","B"};
note\[Breve]number=sharp\[Breve]note\[Breve]delta=flat\[Breve]note\[Breve]delta=natural\[Breve]note\[Breve]delta=sharp\[Breve]signature=note\[Breve]string=ConstantArray[0,12];

setup\[Breve]key[xx_String]:=(
       {key\[Breve]no,key\[Breve]type}=translate\[Breve]key[xx];
       key\[Breve]name=xx;
        );

translate\[Breve]key[xx_String]:=Module[{x=ToUpperCase[xx],mm,twn,keyn,keyt},
   mm=StringTake[x,-3];
   x=StringDrop[x,-3];
   twn=twelve\[Breve]notes;
   p=Position[twn,x]; 
   If[p==={}, 
      twn=twelve\[Breve]notes1;
      p=Position[twn,x]]; 
   If[p==={},Return[]];
   keyn=p[[1,1]];
   keyt=1;
   If[mm != "MAJ" && mm != "MIN",Return[]];
   If[mm=="MIN",
      keyn=keyn+3;
      keyt=-1;
      If[keyn>12,keyn=key\[Breve]no-12];
   ];      
   {keyn,keyt}
];

string\[Breve]representation[n_,key_]:=Module[{tvn,p,ns},
   tvn=If[sharp\[Breve]signature[[key]],twelve\[Breve]notes,twelve\[Breve]notes1];
   ns=tvn[[n]];
   p=ToCharacterCode[ns][[1]]-ToCharacterCode["A"][[1]]+1;
   If[sharp\[Breve]signature[[key]],
      If[natural\[Breve]note\[Breve]delta[[key,p]]!=0,
         If[StringLength[ns]>1,
            ns=StringTake[ns,1],
            ns=ns<>"\[Natural]";
         ];
      ],
      If[natural\[Breve]note\[Breve]delta[[key,p]]!=0,
         If[StringLength[ns]>1,
            ns=StringTake[ns,1],
            ns=ns<>"\[Natural]";
         ];
      ]
   ];
   ns
]

setup\[Breve]all\[Breve]keys[]:=Module[{key,scale,maj\[Breve]scale,flat,flat\[Breve]pos,sharp\[Breve]pos,sscale,next\[Breve]white,tt},
   next\[Breve]white[x_String]:=FromCharacterCode[ToCharacterCode[x]+1]/."H"->"A";   
   key=1;
   While[key <= 12,
      flat=False;
      tt=Join[twelve\[Breve]notes,twelve\[Breve]notes];
      maj\[Breve]scale={0,2,4,5,7,9,11};
      scale=Map[tt[[#+key]]&,maj\[Breve]scale];
      If[Length[Union[Map[StringTake[#,1]&,scale]]]<7,
         flat=True;
         scale=Map[If[StringLength[#]>1,next\[Breve]white[StringTake[#,1]]<>"\[Flat]",#]&,scale];
      ];
      sscale=Sort[scale];
      sharp\[Breve]signature[[key]]=(!flat);
      If[flat,
         note\[Breve]number[[key]]=MapThread[(#1-StringLength[#2]+1)&,{RotateRight[maj\[Breve]scale,2],sscale}]+60;
         flat\[Breve]pos=Map[(StringLength[#]-1)&,sscale];
         flat\[Breve]note\[Breve]delta[[key]]={-1,-1,0,-1,-1,0,-1}+flat\[Breve]pos;
         sharp\[Breve]note\[Breve]delta[[key]]={1,0,1,1,0,1,1}+flat\[Breve]pos;
         natural\[Breve]note\[Breve]delta[[key]]=flat\[Breve]pos,
         (*natural\[Breve]note\[Breve]delta[[key]]=flat\[Breve]pos,*)
         
         note\[Breve]number[[key]]=MapThread[(#1+StringLength[#2]-1)&,{RotateRight[maj\[Breve]scale,2],sscale}]+60;
         sharp\[Breve]pos=Map[(StringLength[#]-1)&,sscale];
         sharp\[Breve]note\[Breve]delta[[key]]={1,0,1,1,0,1,1}-sharp\[Breve]pos;
         flat\[Breve]note\[Breve]delta[[key]]={-1,-1,0,-1,-1,0,-1}+sharp\[Breve]pos;
         natural\[Breve]note\[Breve]delta[[key]]=-sharp\[Breve]pos;
      ];
      note\[Breve]string[[key]]=Table[string\[Breve]representation[j,key],{j,1,12}];
      key++;
   ];
];


SetAttributes[parse\[Breve]to\[Breve]note\[Breve]number,HoldFirst];
parse\[Breve]to\[Breve]note\[Breve]number[str_,key_]:=Module[{nn,nn1},
   String\[Breve]Parse[str,CharacterRange["A","G"]|CharacterRange["a","g"]];
   nn=ToCharacterCode[ToUpperCase[parsed\[Breve]result]][[1]]-64;
   nn1=note\[Breve]number[[key,nn]];
   While[String\[Breve]Parse[str,"\[Sharp]" | "\[Flat]" | "\[Natural]" | DigitCharacter],       
      Switch[parsed\[Breve]result,
         "\[Sharp]", nn1=nn1+sharp\[Breve]note\[Breve]delta[[key,nn]],
         "\[Flat]",nn1=nn1+flat\[Breve]note\[Breve]delta[[key,nn]],
         "\[Natural]",nn1=nn1+natural\[Breve]note\[Breve]delta[[key,nn]],
         _,nn1=nn1+12*(First[ToCharacterCode[parsed\[Breve]result]]-First[ToCharacterCode["4"]]);
      ];
   ];
   nn1
];



get\[Breve]note\[Breve]number\[Breve]string[nn_,key_]:=Module[{f=nn,ttn},
   oct=IntegerPart[f/12];
   note\[Breve]string[[key,f-12*oct+1]]<>FromCharacterCode[oct-1+ToCharacterCode["0"]]
];


(* ::Section:: *)
(*Tempo*)


time\[Breve]rate=Piecewise[{{1,m\[Breve]time>= -1}}];
new\[Breve]tempo[tempo_]:=Module[{},
   time\[Breve]rate=Insert[time\[Breve]rate,{1/beat\[Breve]note*60/tempo,m\[Breve]time>=start\[Breve]time},{1,1}]
];

Set\[Breve]Time\[Breve]Rate\[Breve]Function[fn_]:=(time\[Breve]map\[Breve]interrupt\[Breve]function=fn);
Time\[Breve]Rate[x_]:=(time\[Breve]rate/.m\[Breve]time->x);
Set\[Breve]Time\[Breve]Rate[fn_,x_]:=(time\[Breve]rate=fn/.x->m\[Breve]time);


(* ::Section:: *)
(*Control Palette*)


musical\[Breve]options[]:=(
change\[Breve]generation\[Breve]mode[a_,_]:=(generation\[Breve]mode=a;SetOptions[Render\[Breve]Music,"Result"->a]);
generation\[Breve]mode="Sound";Row[{"Generation mode: ",PopupMenu[Dynamic[generation\[Breve]mode,change\[Breve]generation\[Breve]mode],{"Sound","MusicExpression","SoundObject"}]}]
);


main\[Breve]palette\[Breve]title="Music control panel";
print\[Breve]bar\[Breve]locations=False;
first\[Breve]bar=1;
last\[Breve]bar=\[Infinity];
Window\[Breve]Title[nb_]:=("WindowTitle"/.NotebookInformation[nb]);

SetAttributes[toggle\[Breve]button,HoldFirst];
toggle\[Breve]button[var_,txt_]:=(var=False;Button[Dynamic[Style[txt,If[var,Red,Black]]],var=!var]);

Create\[Breve]Palette[]:=Module[{tmp,padding},
   tmp=Select[Notebooks[],
      (Window\[Breve]Title[#]==main\[Breve]palette\[Breve]title)&];
   If[tmp =!={},
      NotebookClose[tmp[[1]]];
      ];
   padding=Button["",Null,Enabled->False];
   If[! ValueQ[step\[Breve]counter],step\[Breve]counter=0];
   main\[Breve]palette=CreatePalette[Row[{"  ",Column[{" ",
           Tooltip[musical\[Breve]options[],
                        "Select behaviour of Render\[Breve]Music"],

                        Tooltip[Row[{Checkbox[Dynamic[print\[Breve]bar\[Breve]locations]],"Print timing info"}],
                        "Print the time (in whole notes) of each bar"],
                         Button["Playback range",set\[Breve]playback\[Breve]range[]],               
                          " "},Spacings->1],"   "}],             
         WindowFrameElements->{"CloseBox"},
         WindowElements->{(*"StatusArea",*)"MenuBar"},Selectable->True,
         WindowTitle->main\[Breve]palette\[Breve]title];
   NotebookSave[main\[Breve]palette,palette\[Breve]path];
   ];


set\[Breve]playback\[Breve]range[]:=CreateDialog[Column[{
         Row[{"Bar: ",
               InputField[Dynamic[first\[Breve]bar],Number,ImageSize->50]," to ", 
               InputField[Dynamic[last\[Breve]bar],Number,ImageSize->50]}],
         Row[{Button["Full range",first\[Breve]bar=1;last\[Breve]bar=10000],"  ",DefaultButton[]}],
          " "
      }
   ],WindowTitle->"Playback range"
];


Create\[Breve]Palette[];


(* ::Section:: *)
(*Rendering*)


instruments={"Piano"};
setup\[Breve]all\[Breve]keys[];

initialise[]:=Module[{},
time\[Breve]tate=Piecewise[{{1,mtm>= -1}}];
   bar\[Breve]length=1;
(* tempo specified in quarter notes*)
   Clear[bar\[Breve]times];
   beat\[Breve]note=1/4;
   pulse\[Breve]time=-1;
   current\[Breve]bar=0;
   start\[Breve]found=False;
   end\[Breve]found=False;
   playback\[Breve]start=0;
   playback\[Breve]end=\[Infinity];
   freq\[Breve]eps=0;
   length\[Breve]scale=1;
   start\[Breve]t=0;
   start\[Breve]time=0.0;
   setup\[Breve]key["CMAJ"];
   current\[Breve]instrument=1;
   end\[Breve]t=15;
   total\[Breve]t=end\[Breve]t-start\[Breve]t;
   new\[Breve]tempo[120];    
   all\[Breve]notes={};
   volume=1;
];

initialise[];


To\[Breve]Music\[Breve]String[x_seq]:=Module[{tmp},
     tmp=tms[x];
     If[StringTake[tmp,1]=="(" && StringTake[tmp,-1]==")",
       tmp=StringDrop[StringDrop[tmp,1],-1]];
     StringReplace[tmp,{"( "->"("," )"->")"}]
];

tms[x_String]:=x;
tms[seq[x___]]:="( "<>StringJoin@@Riffle[Map[tms,{x}]," "]<>" )";
tms[par[x___]]:=StringJoin@@Riffle[Map[tms,{x}],"+"];
tms[Hold[fn_][x___]]:=SymbolName[fn]<>"("<>StringJoin@@Riffle[Map[tms,{x}],","]<>")";
tms[x_?NumericQ]:=ToString[x];
tms[x_]:=Error["Unable to convert "<>ToString[x]];


list\[Breve]flatten[x_List]:=Sequence@@x;
list\[Breve]flatten[x_]:=x;


volume\[Breve]multiplier=0.6;

time\[Breve]map[x_]:=(time\[Breve]rate\[Breve]int/.m\[Breve]time->x);
Options[Render\[Breve]Music]={"Result"->"Sound"};
Render\[Breve]Music[expr_,OptionsPattern[Render\[Breve]Music]]:=Module[{required\[Breve]result,ss,min\[Breve]t,tmp,instr,vol,t1,from,to,pbs,synth},
   required\[Breve]result=OptionValue["Result"];
   initialise[];   
   tmp=expr;
   If[StringQ[tmp],tmp=Parse\[Breve]Music[tmp]];
   If[required\[Breve]result=="MusicExpression",Return[tmp]];
   rm1[tmp];  
   print\[Breve]timing\[Breve]info[];
   If[ValueQ[time\[Breve]rate\[Breve]interrupt\[Breve]function],time\[Breve]rate\[Breve]interrupt\[Breve]function[]];
   from=0;
   to=\[Infinity];
   ss=Select[all\[Breve]notes,(#[[2]]>=from && #[[2]]<= to)&];   
   time\[Breve]rate\[Breve]int=Integrate[time\[Breve]rate,m\[Breve]time];
   pbs=time\[Breve]map[playback\[Breve]start];
   min\[Breve]t=Min[Map[#[[2]]&,ss]];
   synth=Map[With[{name=#[[1]],fn=#[[2]]},nn:SoundNote[n_,{start_,end_},name,___]:>
   list\[Breve]flatten[fn[n+60,start,end,Options[nn,SoundVolume][[1,2]]]]]&,synthetic\[Breve]sound\[Breve]table];
   ss=Map[(tmp=#;tmp[[2]]=tmp[[2]]-min\[Breve]t;tmp)&,ss];   
   ss=ss/.note[n_,s_,l_,vol_,instr_,___]:>If[s>=playback\[Breve]start && s<= playback\[Breve]end,SoundNote[n-60,{time\[Breve]map[s]-pbs,time\[Breve]map[s+l]-pbs}(*/beat\[Breve]note*60/tempo*),instruments[[instr]],SoundVolume->(volume\[Breve]multiplier*2^(vol-1))],Sequence@@{}];
   (* Caution - if we use more SoundNote features, this might need to change *)
   ss=ss/.synth;
   ss=Sound[ss];
   MusicGenerator`Last\[Breve]Sound=ss;
    If[required\[Breve]result=="SoundObject",Return[ss]];   
   EmitSound[ss];
   tmp=Map[#[[2,2]]&,ss[[1]]];
   If[Length[tmp]>0,tmp=Max[tmp]];
   If[tmp>0,Pause[tmp]];
];


print\[Breve]timing\[Breve]info[]:=Module[{},
    Print["No. whole notes=",start\[Breve]time];
]


rm1[x:Hold[fn_][___]]:=Module[{ans},
   ans=fn@@x;
   If[ListQ[ans],
      all\[Breve]notes=Join[all\[Breve]notes,ans],
      rm1[ans]
   ]
];

Bar[n_]:=(
   current\[Breve]bar=n;
   If[current\[Breve]bar==first\[Breve]bar && ! start\[Breve]found,
      playback\[Breve]start=start\[Breve]time;
      start\[Breve]found=True
   ];
   If[current\[Breve]bar==last\[Breve]bar+1 && ! end\[Breve]found,
      playback\[Breve]end=start\[Breve]time;
      end\[Breve]found=True;
   ];
   pulse\[Breve]time=start\[Breve]time;
   bar\[Breve]times[n]=start\[Breve]time;
   If[print\[Breve]bar\[Breve]locations,Print["Bar ",n," at ",start\[Breve]time]]
);
(*bar[_]:=(pulse\[Breve]time=start\[Breve]time;);*)

(*SetAttributes[seq,Flat];*)
rm1[seq[x___]]:=Module[{ll={x}},

   Block[{current\[Breve]instrument=current\[Breve]instrument,key\[Breve]no=key\[Breve]no,key\[Breve]type=key\[Breve]type,key\[Breve]name=key\[Breve]name,
         volume=volume},
      Map[rm1,ll];
   ];
];

rm1[par[x___]]:=Module[{ll={x},s=start\[Breve]time,s1,k},
   s1=s;
   k=1;
   While[k <= Length[ll],
      start\[Breve]time=s;
      rm1[ll[[k]]];
      s1=Max[s1,start\[Breve]time];
      k++;
   ];
   start\[Breve]time=s1;
];

(*\[LightBulb]1*)rm1[s_String]:=Module[{str=ToUpperCase[s]<>"         ",note\[Breve]number,ch,
      instrument,nl,nl\[Breve]eps,tmp,nn,vol,pos},
   nl=length\[Breve]scale/4;
   instrument=current\[Breve]instrument;
   ch=StringTake[str,1];
   this\[Breve]chord="";
   Which[
      (* Parsing base note *)
      StringMatchQ[StringTake[str,1],CharacterRange["A","G"]],      
      note\[Breve]number=parse\[Breve]to\[Breve]note\[Breve]number[str,key\[Breve]no],
      (*str=StringReplace[str,{"\[Sharp]"->"","\[Flat]"->"","\[Natural]"->""}],*)
      
      (* Parsing rest *)
      ch=="R",
      str=StringDrop[str,1];
      note\[Breve]number=-1,
      
      (* Parsing tempo *)
      ch=="T",
      str=StringDrop[str,1];
      If[! String\[Breve]Parse[str,DigitCharacter..],Error["Missing number following 'T'"]];
      new\[Breve]tempo[ToExpression[parsed\[Breve]result]];       
      Return[], 
      
      ch=="V",
      str=StringDrop[str,1];
      If[! String\[Breve]Parse[str,"F"|"P"],Error["V should be followed by 'F' or 'P'"]];
      If[parsed\[Breve]result=="F",
         volume++,
         volume--
      ];
      Return[], 
      
      (* Parsing current instrument *)
      ch=="I",
      str=StringDrop[str,1];
      If[! String\[Breve]Parse[str,DigitCharacter..],Error["Missing number following 'I'"]];
      current\[Breve]instrument=ToExpression[parsed\[Breve]result];       
      Return[], 
      
      (* Parsing scale *)
      ch=="K",
      str=StringDrop[str,1];
      If[!String\[Breve]Parse[str,CharacterRange["A","G"]~~("\[Sharp]"|"\[Flat]"|"")~~("MAJ"|"MIN")],Error["Invalid key name following 'K'"]];
      setup\[Breve]key[parsed\[Breve]result];
      Return[],
      True,
      Error["Unexpected '"<>ch<>"' encountered"]
   ];
   
 nl\[Breve]eps=0;
   While[True,
      (* possible explicit length *)
      ch=StringTake[str,1];
      str=StringDrop[str,1];
      Switch[ch,
      " ",Break[],
      "*",
      If[! String\[Breve]Parse[str,DigitCharacter..~~"/"~~DigitCharacter..],Error["Invalid use of '*'"]];
         (*tmp=StringCases[str,StartOfString~~"*"~~DigitCharacter..~~"/"~~DigitCharacter..];*)
         nl=ToExpression[parsed\[Breve]result],
      "W",nl=length\[Breve]scale,
      "H",nl=length\[Breve]scale/2,
      "Q",nl=length\[Breve]scale/4,
      "I",nl=length\[Breve]scale/8,
       "X",nl=length\[Breve]scale/16,
      "O",nl=length\[Breve]scale/32,
      ".",nl=nl*(3/2),
      "/",
       If[String\[Breve]Parse[str,"3"],nl=nl/3,
         If[String\[Breve]Parse[str,"5"],nl=nl/5];
      ],
      (* possible instrument selection *)
      "'",
      String\[Breve]Parse[str,"'"...];
        tmp=StringLength[parsed\[Breve]result]+1;
        instrument=Mod[current\[Breve]instrument+tmp,Length[instruments],1],
      (* possible legato/stoccato indicators *)
      "S",nl\[Breve]eps=-nl/3,
      "L",nl\[Breve]eps=nl/3,
      _,Error["Note string "<>s<>" not recognised"]
      ];      
   ];
   
   vol=volume;
   If[pulse\[Breve]time==start\[Breve]time,vol=vol*1.9];
   If[note\[Breve]number>=0,
      tmp={note[note\[Breve]number+freq\[Breve]eps,start\[Breve]time,nl+nl\[Breve]eps,vol,instrument]};   
      all\[Breve]notes=Join[all\[Breve]notes,tmp];
   ];
   start\[Breve]time=start\[Breve]time+nl;
];


(* ::Section:: *)
(*Music functions*)


rep[x_,n_Integer]:=seq@@Table[x,{n}];


rall[x_]:=rall[x,0.5];
rall[x_,slow_]:=Module[{t1,t2,var,tmp,ff,gg,tt,init\[Breve]v,v\[Breve]t},
   t1=start\[Breve]time;
   rm1[x];
   t2=start\[Breve]time;
   init\[Breve]v=Time\[Breve]Rate[t1];
   v\[Breve]t=init\[Breve]v (t2-t)/(t2-t1)+init\[Breve]v/slow*(t-t1)/(t2-t1);
   time\[Breve]rate=Insert[Insert[time\[Breve]rate,
{v\[Breve]t/.t->m\[Breve]time,m\[Breve]time>t1},{1,1}],
{init\[Breve]v,m\[Breve]time>t2},{1,1}];
];


trp[x_,n_:3]:=do\[Breve]tr[x,n,1];
trm[x_,n_:3]:=do\[Breve]tr[x,n,-1];

do\[Breve]tr[x_,n_,sg_]:=Module[{nt,ans={},j,t,nl},
   Block[{freq\[Breve]eps=freq\[Breve]eps,length\[Breve]scale=length\[Breve]scale/n},   
      j=1;
      While[j <= n,
         freq\[Breve]eps=freq\[Breve]eps+sg*BitAnd[j-1,1];
         rm1[x];
         freq\[Breve]eps=freq\[Breve]eps-sg*BitAnd[j-1,1];      
         j++;
      ];
   ];
];


(* ::Section:: *)
(*Music manipulation functions*)


Note\[Breve]StringQ[x_String]:=StringMatchQ[x,(CharacterRange["A","G"]|CharacterRange["a","g"])~~
        (DigitCharacter| "\[Sharp]" | "\[Flat]" | "\[Natural]")~~___];


(*Note\[Breve]Pattern[]:=(x_String?Note\[Breve]StringQ);*)


Transpose\[Breve]Note[str_,n_Integer,key\[Breve]name_]:=Module[{tmp=str,nv,key},
   key=First[translate\[Breve]key[key\[Breve]name]];
   nv=parse\[Breve]to\[Breve]note\[Breve]number[tmp,key]+n;
   get\[Breve]note\[Breve]number\[Breve]string[nv,key]<>tmp
];

Invert\[Breve]Note\[Breve]About[str_,pivot_,key\[Breve]name_]:=Module[{tmp=str,nv,key,pv,tmp1},
   key=First[translate\[Breve]key[key\[Breve]name]];
   nv=parse\[Breve]to\[Breve]note\[Breve]number[tmp,key];
   tmp1=pivot;
   pv=parse\[Breve]to\[Breve]note\[Breve]number[tmp1,key];
   nv=pv-nv+60;
   get\[Breve]note\[Breve]number\[Breve]string[nv,key]<>tmp
];




(* ::Section:: *)
(*Named chord functions*)


add9[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,14}];
dim7[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,6,9}];
min9[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,7,10,14}];
maj9[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,11,14}];
dom9[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,10,14}];
min6[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,7,9}];
maj6[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,9}];
sus2[x_]:=do\[Breve]named\[Breve]chord[x,{0,2,7}];
sus4[x_]:=do\[Breve]named\[Breve]chord[x,{0,5,7}];
min7[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,7,10}];
maj7[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,11}];
dom7[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7,10}];
dim[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,6}];
aug[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,8}];
min[x_]:=do\[Breve]named\[Breve]chord[x,{0,3,7}];
maj[x_]:=do\[Breve]named\[Breve]chord[x,{0,4,7}];

do\[Breve]named\[Breve]chord[x_,lst_]:=Module[{j},
   j=1;
   While[j <= Length[lst],
      Block[{freq\[Breve]eps=freq\[Breve]eps,start\[Breve]time=start\[Breve]time},
         freq\[Breve]eps=freq\[Breve]eps+lst[[j]];
         rm1[x];
      ];
      j++;
   ];
];


(* ::Section:: *)
(*Synthetic sounds*)


synthetic\[Breve]sound\[Breve]table={};

Add\[Breve]Synthetic\[Breve]Sound[name_,fn_]:=Module[{},
   synthetic\[Breve]sound\[Breve]table=Append[DeleteCases[synthetic\[Breve]sound\[Breve]table,{name,_}],{name,fn}];
   Length[synthetic\[Breve]sound\[Breve]table];
];


Frequency\[Breve]From\[Breve]Note[n_Integer]:=13.75*2^(n/12);
Frequency\[Breve]From\[Breve]Note[s_String,key_String:"CMAJ"]:=Module[{str=s},
   Frequency\[Breve]From\[Breve]Note[parse\[Breve]to\[Breve]note\[Breve]number[str,First[translate\[Breve]key[key]]]]
];


(* ::Section:: *)
(*Import of musicXML*)


pick\[Breve]bars[mxl_,from_,to_]:=
Cases[mxl,XMLElement["measure",{"number"->n_},___]/;
   (from<=ToExpression[n]<=to),\[Infinity]];

get\[Breve]elem[ll_List,type_,pos_]:=Module[{tmp},
   tmp=Cases[ll,XMLElement[type,___]];
   If[tmp==={},Return[$Failed]];
   Part[tmp[[1]],Sequence@@pos]
];

get\[Breve]elem\[Breve]value[ll_List,type_,pos_]:=Module[{tmp},
   tmp=Cases[ll,XMLElement[type,___]];
   If[tmp==={},Return[$Failed]];
   ToExpression[Part[tmp[[1]],Sequence@@pos]]
];

StringForm\[Breve]String[x___]:=ToString[StringForm[x]];

nl\[Breve]rec=False;
note\[Breve]length[1]="w";
note\[Breve]length[1/2]="h";
note\[Breve]length[1/4]="q";
note\[Breve]length[1/8]="i";
note\[Breve]length[1/16]="x";
note\[Breve]length[1/32]="o";
note\[Breve]length[3/4]="h.";
note\[Breve]length[3/8]="q.";
note\[Breve]length[3/16]="i.";
note\[Breve]length[3/32]="x.";
note\[Breve]length[x_]:=Module[{s},
   nl\[Breve]rec=True;
   s=note\[Breve]length[3 x];
   nl\[Breve]rec=False;
   If[StringQ[s],Return[s<>"/3"]];
   nl\[Breve]rec=True;
   s=note\[Breve]length[5 x];
   nl\[Breve]rec=False;
   If[StringQ[s],Return[s<>"/5"]];
   "*"<>ToString[Numerator[x]]<>"/"<>ToString[Denominator[x]]
]/;(! nl\[Breve]rec);


Options[Import\[Breve]MusicXML]={"BarRange"->{1,\[Infinity]},"Verbose"->False};
Import\[Breve]MusicXML[file_,key\[Breve]name_String:"CMaj",OptionsPattern[Import\[Breve]MusicXML]]:=Module[{ss,nn,ne,dt,time,last\[Breve]time,bar\[Breve]time,
      tmp,tmp1,sf,k,chord,result,this\[Breve]bar,first\[Breve]bar,last\[Breve]bar,all\[Breve]bars,this\[Breve]b,
      key,sf\[Breve]lookup,ans,verbose\[Breve]mode},
   sf\[Breve]lookup={"\[Flat]","","\[Sharp]"};
   key=key=First[translate\[Breve]key[key\[Breve]name]];
   result="K"<>key\[Breve]name<>" ";
   verbose\[Breve]mode=OptionValue["Verbose"];
   ss=Import[file,"XML","ReadDTD"->False];
   divisions\[Breve]per\[Breve]whole\[Breve]note=4*Cases[ss,XMLElement["divisions",_,{x_}]:>ToExpression[x],\[Infinity]][[1]];
   time=0;
   last\[Breve]time=0;
   tmp=Cases[ss,XMLElement["part",___],\[Infinity]];
   If[Length[tmp]==0,Return[$Failed]];
   (*all\[Breve]bars=tmp[[1,3]];*)
   all\[Breve]bars=Cases[tmp,XMLElement["measure",___],\[Infinity]];
   tmp=OptionValue["BarRange"];
   If[! ListQ[tmp],tmp={1,tmp}];
   first\[Breve]bar=Max[1,tmp[[1]]];
   last\[Breve]bar=Min[Length[all\[Breve]bars],tmp[[2]]];
      this\[Breve]bar=first\[Breve]bar;
   While[this\[Breve]bar <= last\[Breve]bar,
      this\[Breve]b="";
      bar\[Breve]time=time;
      (*Print["Bar: ",this\[Breve]bar];*)
      this\[Breve]b=this\[Breve]b<>StringForm\[Breve]String["Bar(``) ( ",this\[Breve]bar];
      (* nn=pick\[Breve]bars[ss,this\[Breve]bar,this\[Breve]bar][[1,3]];  *)
      nn=all\[Breve]bars[[this\[Breve]bar,3]];    
      k=1;
      While[k<=Length[nn],
         Switch[nn[[k,1]],
            "note",
            ne=nn[[k,3]];
            sf=0;
            tmp=get\[Breve]elem\[Breve]value[ne,"alter",{3}];
            If[tmp=!=$Failed,sf=tmp];
            chord=get\[Breve]elem[ne,"chord",{3}];
            If[chord=!=$Failed,time=last\[Breve]time];
            tmp=get\[Breve]elem[ne,"pitch",{3}];
            ans=If[tmp===$Failed,"R",

               tmp1=get\[Breve]elem\[Breve]value[tmp,"alter",{3}];
               If[tmp1=!=$Failed,sf=tmp1];
               get\[Breve]elem[tmp,"step",{3,1}]<>sf\[Breve]lookup[[sf+2]]<>get\[Breve]elem[tmp,"octave",{3,1}]
];
            If[key!=1 && ans!="R",
            ans=get\[Breve]note\[Breve]number\[Breve]string[parse\[Breve]to\[Breve]note\[Breve]number[ans,1],key];
            ];
            dt=get\[Breve]elem\[Breve]value[ne,"duration",{3,1}]/divisions\[Breve]per\[Breve]whole\[Breve]note;
            (*Print[{ans,"time"->time,"len"->dt}];*)
            this\[Breve]b=this\[Breve]b<>If[chord===$Failed," ","+"]<>ans<>note\[Breve]length[dt];
            last\[Breve]time=time;
            time=time+dt,
            "backup",
            dt=get\[Breve]elem\[Breve]value[nn[[k,3]],"duration",{3,1}];
            (* Print["Backup ",dt];    *)        
            time=time-dt/divisions\[Breve]per\[Breve]whole\[Breve]note;
            this\[Breve]b=this\[Breve]b<>" )+( ";
            If[time > bar\[Breve]time,
               this\[Breve]b=this\[Breve]b<>"r"<>note\[Breve]length[time-bar\[Breve]time]<>" ";
            ];            
            last\[Breve]time=time,
            "attributes",
            ne=nn[[k,3]];
            tmp=get\[Breve]elem\[Breve]value[ne,"divisions",{3,1}];
            If[tmp=!=$Failed,
               divisions\[Breve]per\[Breve]whole\[Breve]note=4*tmp;
            ];
            tmp=get\[Breve]elem\[Breve]value[ne,"key",{3,1,3,1}];            
            If[tmp=!=$Failed && verbose\[Breve]mode,
               Print["Key: ",tmp];
            ];
            tmp=get\[Breve]elem\[Breve]value[ne,"beats",{3,1}];
            If[tmp=!=$Failed,
               beats=tmp;
            ];
            tmp=get\[Breve]elem\[Breve]value[ne,"beat-type",{3,1}];
            If[tmp=!=$Failed,
               beats\[Breve]type=tmp;
            ],
            "direction",Null,
            _,
            If[verbose\[Breve]mode,Print["Unexpected ",nn[[k]]]];
         ];
         k++;
      ];
      this\[Breve]b=this\[Breve]b<>")";
      result=result<>this\[Breve]b;
      this\[Breve]bar++;
   ];   
   result
];


(* ::Section:: *)
(*Package end*)


End[];


EndPackage[];
