(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[      9767,        311]
NotebookOptionsPosition[      7097,        226]
NotebookOutlinePosition[      8993,        283]
CellTagsIndexPosition[      8912,        278]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Sqrt \[Dash] additional help", "Chapter",
 CellChangeTimes->{{3.4422352966623373`*^9, 3.442235300928472*^9}, {
   3.44731239346875*^9, 3.447312399453125*^9}, {3.447390701375*^9, 
   3.447390712625*^9}, {3.447390810796875*^9, 3.4473908173125*^9}, {
   3.4523472274375*^9, 3.452347231625*^9}, {3.4996253268141565`*^9, 
   3.4996253368297815`*^9}, {3.4996255731735315`*^9, 3.4996255813610315`*^9}, 
   3.499889488421875*^9, 3.50018953578125*^9, {3.500308735703125*^9, 
   3.500308736359375*^9}, {3.5003974425*^9, 3.500397445609375*^9}, {
   3.50451034953125*^9, 3.504510361046875*^9}, {3.504511482390625*^9, 
   3.504511486359375*^9}, {3.5062882946875*^9, 3.506288295828125*^9}},
 CellTags->"Define\[Breve]Macro"],

Cell[TextData[{
 StyleBox["Author :",
  FontWeight->"Bold"],
 " ",
 ButtonBox["David Bailey",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.dbaileyconsultancy.co.uk"], None},
  ButtonNote->"http://www.dbaileyconsultancy.co.uk"]
}], "Text",
 CellChangeTimes->{{3.50453037203125*^9, 3.504530387828125*^9}, {
  3.504530423171875*^9, 3.504530423171875*^9}}],

Cell["\<\
Users are sometimes puzzled that Sqrt doesn't always seem to evaluate, for \
example:\
\>", "Text",
 CellChangeTimes->{{3.504510377*^9, 3.504510508359375*^9}, 
   3.50451149403125*^9, {3.5062883199375*^9, 3.506288359515625*^9}, {
   3.506288461375*^9, 3.506288464609375*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Sqrt", "[", "4.0", "]"}], ",", 
   RowBox[{"Sqrt", "[", "4", "]"}], ",", 
   RowBox[{"Sqrt", "[", "3", "]"}], ",", 
   RowBox[{"Sqrt", "[", 
    RowBox[{"2", "/", "3"}], "]"}], ",", 
   RowBox[{"Sqrt", "[", 
    RowBox[{"2.0", "/", "3"}], "]"}]}], "}"}]], "Input",
 CellChangeTimes->{{3.506288384296875*^9, 3.506288448046875*^9}, {
  3.506288552171875*^9, 3.506288572546875*^9}},
 CellLabel->"In[2]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2.`", ",", "2", ",", 
   SqrtBox["3"], ",", 
   SqrtBox[
    FractionBox["2", "3"]], ",", "0.816496580927726`"}], "}"}]], "Output",
 CellChangeTimes->{3.5062884489375*^9, 3.5062885743125*^9},
 CellLabel->"Out[2]="]
}, Open  ]],

Cell[TextData[{
 "This is because ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " never converts exact quantities into approximate ones, unless explicitly \
requested to do so \[Dash] e.g. by using N. Expressions involving decimal \
numbers are already considered to be approximate, and so Sqrt operated on \
them immediately. Alternatively, N will always force evaluation:"
}], "Text",
 CellChangeTimes->{{3.506288489421875*^9, 3.5062885428125*^9}, {
  3.506288590328125*^9, 3.506288603171875*^9}, {3.50628864784375*^9, 
  3.506288745703125*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Sqrt", "[", "4.0", "]"}], ",", 
    RowBox[{"Sqrt", "[", "4", "]"}], ",", 
    RowBox[{"Sqrt", "[", "3", "]"}], ",", 
    RowBox[{"Sqrt", "[", 
     RowBox[{"2", "/", "3"}], "]"}], ",", 
    RowBox[{"Sqrt", "[", 
     RowBox[{"2.0", "/", "3"}], "]"}]}], "}"}], "//", "N"}]], "Input",
 CellChangeTimes->{{3.506288750828125*^9, 3.50628875178125*^9}},
 CellLabel->"In[3]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "2.`", ",", "2.`", ",", "1.7320508075688772`", ",", "0.816496580927726`", 
   ",", "0.816496580927726`"}], "}"}]], "Output",
 CellChangeTimes->{3.50628875371875*^9},
 CellLabel->"Out[3]="]
}, Open  ]],

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " does this for a reason \[Dash] because the approximation of exact \
expressions would be totally inapprpopriate in algebraic/calculus \
applications, for example:"
}], "Text",
 CellChangeTimes->{{3.5064445560625*^9, 3.506444595328125*^9}, {
  3.5064446326875*^9, 3.5064446975*^9}, {3.506444777234375*^9, 
  3.50644479003125*^9}, {3.506444849421875*^9, 3.50644485275*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Series", "[", 
  RowBox[{
   RowBox[{"Sqrt", "[", 
    RowBox[{"2", "-", 
     RowBox[{"x", "^", "2"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0", ",", "6"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.506444702875*^9, 3.506444745796875*^9}, {
  3.506444805859375*^9, 3.50644482609375*^9}},
 CellLabel->"In[4]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{
   SqrtBox["2"], "-", 
   FractionBox[
    SuperscriptBox["x", "2"], 
    RowBox[{"2", " ", 
     SqrtBox["2"]}]], "-", 
   FractionBox[
    SuperscriptBox["x", "4"], 
    RowBox[{"16", " ", 
     SqrtBox["2"]}]], "-", 
   FractionBox[
    SuperscriptBox["x", "6"], 
    RowBox[{"64", " ", 
     SqrtBox["2"]}]], "+", 
   InterpretationBox[
    SuperscriptBox[
     RowBox[{"O", "[", "x", "]"}], "7"],
    SeriesData[$CellContext`x, 0, {}, 0, 7, 1],
    Editable->False]}],
  SeriesData[$CellContext`x, 0, {
   2^Rational[1, 2], 0, Rational[-1, 2] 2^Rational[-1, 2], 0, 
    Rational[-1, 16] 2^Rational[-1, 2], 0, Rational[-1, 64] 
    2^Rational[-1, 2]}, 0, 7, 1],
  Editable->False]], "Output",
 CellChangeTimes->{{3.506444733453125*^9, 3.5064447465*^9}, {
  3.50644480890625*^9, 3.506444827546875*^9}},
 CellLabel->"Out[4]="]
}, Open  ]],

Cell["Some algebraic expressions also fail to evaluate:", "Text",
 CellChangeTimes->{{3.50628882684375*^9, 3.506288892296875*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Sqrt", "[", 
  RowBox[{"a", "^", "2"}], "]"}]], "Input",
 CellChangeTimes->{{3.506288777421875*^9, 3.5062887834375*^9}},
 CellLabel->"In[4]:="],

Cell[BoxData[
 SqrtBox[
  SuperscriptBox["a", "2"]]], "Output",
 CellChangeTimes->{3.5062887873125*^9},
 CellLabel->"Out[4]="]
}, Open  ]],

Cell["\<\
The problem is that there are two possible answers \[Dash] a or -a. The \
solution is to use PowerExpand, assuming that you are not concerned by the \
alternative answer.\
\>", "Text",
 CellChangeTimes->{{3.506288898359375*^9, 3.506289002796875*^9}, {
  3.5062890450625*^9, 3.50628910740625*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Sqrt", "[", 
   RowBox[{"a", "^", "2"}], "]"}], "//", "PowerExpand"}]], "Input",
 CellChangeTimes->{{3.506288977421875*^9, 3.506288981625*^9}},
 CellLabel->"In[5]:="],

Cell[BoxData["a"], "Output",
 CellChangeTimes->{3.50628898325*^9},
 CellLabel->"Out[5]="]
}, Open  ]],

Cell[TextData[{
 "Similar problems of avoiding approximation and multiple values occur with \
many other ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions."
}], "Text",
 CellChangeTimes->{{3.506289114890625*^9, 3.50628914021875*^9}, {
  3.506444912671875*^9, 3.506445000375*^9}}]
}, Open  ]]
},
Saveable->True,
Selectable->True,
WindowSize->{717, 741},
WindowMargins->{{Automatic, 151}, {Automatic, 66}},
WindowFrame->"Normal",
Visible->True,
DockedCells->FEPrivate`FrontEndResource["FEExpressions", "HelpViewerToolbar"],
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PageHeaders->{{Inherited, Inherited, 
   Cell[
    TextData[{"The Super Widget Package"}], "PageHeader"]}, {
   Cell[
    TextData[{"The Super Widget Package"}], "PageHeader"], Inherited, 
   Inherited}},
PageFooters->{{Inherited, 
   Cell[
    TextData[{"\[Copyright]2006 David Bailey"}], "PageFooter"], Inherited}, {
  Inherited, 
   Cell[
    TextData[{"\[Copyright]2006 David Bailey"}], "PageFooter"], Inherited}},
PrivateNotebookOptions->{"ShortBoxForm"->False},
ShowCellBracket->True,
ShowSelection->True,
TaggingRules->{
 "ModificationHighlight" -> False, 
  "Metadata" -> {
   "context" -> "DebugTrace`", "keywords" -> {}, "index" -> True, "label" -> 
    "DebugTrace Package Paclet Symbol", "language" -> "en", "paclet" -> 
    "DebugTrace Package", "status" -> "", "summary" -> 
    "DTLocalValue[n] \[Dash] Returns the value of the most recently created \
local variable of the form n$<number>", "synonyms" -> {}, "title" -> 
    "DTLocalValue", "type" -> "Symbol", "uri" -> 
    "DebugTrace/ref/DTLocalValue"}, "SearchTextTranslated" -> "", 
  "SearchText" -> "", "LinkTrails" -> ""},
ShowCellLabel->False,
PrivateNotebookOptions -> {"NotebookAuthor" -> ""},
Magnification->1.,
FrontEndVersion->"7.0 for Microsoft Windows (64-bit) (February 18, 2009)",
StyleDefinitions->"HelpBrowser.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Define\[Breve]Macro"->{
  Cell[579, 22, 716, 10, 85, "Chapter",
   CellTags->"Define\[Breve]Macro"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Define\[Breve]Macro", 8811, 272}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[579, 22, 716, 10, 85, "Chapter",
 CellTags->"Define\[Breve]Macro"],
Cell[1298, 34, 369, 11, 23, "Text"],
Cell[1670, 47, 285, 6, 23, "Text"],
Cell[CellGroupData[{
Cell[1980, 57, 454, 12, 27, "Input"],
Cell[2437, 71, 254, 7, 49, "Output"]
}, Open  ]],
Cell[2706, 81, 554, 11, 55, "Text"],
Cell[CellGroupData[{
Cell[3285, 96, 436, 12, 27, "Input"],
Cell[3724, 110, 230, 6, 25, "Output"]
}, Open  ]],
Cell[3969, 119, 439, 9, 39, "Text"],
Cell[CellGroupData[{
Cell[4433, 132, 358, 10, 27, "Input"],
Cell[4794, 144, 873, 28, 48, "Output"]
}, Open  ]],
Cell[5682, 175, 130, 1, 23, "Text"],
Cell[CellGroupData[{
Cell[5837, 180, 167, 4, 27, "Input"],
Cell[6007, 186, 126, 4, 33, "Output"]
}, Open  ]],
Cell[6148, 193, 306, 6, 39, "Text"],
Cell[CellGroupData[{
Cell[6479, 203, 201, 5, 27, "Input"],
Cell[6683, 210, 89, 2, 25, "Output"]
}, Open  ]],
Cell[6787, 215, 294, 8, 23, "Text"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
